"""Client utilities for accessing NILM database via HTTP"""

from __future__ import absolute_import
import nilmdb
from nilmdb.printf import *

import time
import sys
import re
import os
import json
import urlparse
import urllib
import urllib2
from urllib2 import urlopen, HTTPError
import pycurl

class NilmCommError(Exception):
    """Base exception for both ClientError and ServerError responses"""
    # Whether to include the body in the __str__ expansion
    show_body = 1
    def __init__(self, code, message, body = None, url = "no URL"):
        Exception.__init__(self, message)
        self.code = code
        self.body = body
        self.url = url
    def __str__(self): # pragma: no cover
        s = sprintf("[%03d] %s (%s)", self.code, self.message, self.url)
        if self.body and self.show_body:
            s += "\n" + self.body
        return s

class ClientError(NilmCommError):
    pass

class ServerError(NilmCommError):
    pass

class MyCurl(object):
    def __init__(self, baseurl = ""):
        """If baseurl is supplied, all other functions that take
        a URL can be given a relative URL instead."""
        self.baseurl = baseurl
        self.url = baseurl
        self.curl = pycurl.Curl()
        self.curl.setopt(pycurl.SSL_VERIFYHOST, 2)
        self.curl.setopt(pycurl.FOLLOWLOCATION, 1)
        self.curl.setopt(pycurl.MAXREDIRS, 5)

    def _setup_url(self, url, params):
        url = urlparse.urljoin(self.baseurl, url)
        if params:
            url = urlparse.urljoin(url, "?" + urllib.urlencode(params, True))
        self.curl.setopt(pycurl.URL, url)
        self.url = url

    def _check_error(self, body = None):
        code = self.curl.getinfo(pycurl.RESPONSE_CODE)
        if code >= 400 and code <= 499:
            raise ClientError(code, "HTTP client error", body, self.url)
        elif code >= 500 and code <= 599: # pragma: no cover
            raise ServerError(code, "HTTP server error", body, self.url)
        elif code != 200: # pragma: no cover
            raise NilmCommError(code, "Unknown error", body, self.url)
        
    def getjson(self, url, params = None):
        """Simple GET that returns JSON string"""
        self._setup_url(url, params)
        self._body = ""
        def body_callback(x):
            self._body += x
        self.curl.setopt(pycurl.WRITEFUNCTION, body_callback)
        try:
            self.curl.perform()
        except pycurl.error as e:
            raise ServerError(500, e[1], self.url)
        self._check_error(self._body)
        return json.loads(self._body)

class Client(object):
    def __init__(self, url):
        self.curl = MyCurl(url)

    def _json_param(self, data):
        """Return compact json-encoded version of parameter"""
        return json.dumps(data, separators=(',',':'))

    def version(self):
        return self.curl.getjson("version")

    def stream_list(self, path = None, layout = None):
        params = {}
        if path is not None:
            params["path"] = path
        if layout is not None:
            params["layout"] = layout
        return self.curl.getjson("stream/list", params)

    def stream_get_metadata(self, path, keys = None):
        params = { "path": path }
        if keys is not None:
            params["key"] = keys
        return self.curl.getjson("stream/get_metadata", params)

    def stream_set_metadata(self, path, data):
        """Set stream metadata from a dictionary, replacing all existing
        metadata."""
        params = {
            "path": path,
            "data": self._json_param(data)
            }
        return self.curl.getjson("stream/set_metadata", params)

    def stream_update_metadata(self, path, data):
        """Update stream metadata from a dictionary"""
        params = {
            "path": path,
            "data": self._json_param(data)
            }
        return self.curl.getjson("stream/update_metadata", params)

    def stream_create(self, path, layout, index = None):
        params = { "path": path,
                   "layout" : layout }
        if index is not None:
            params["index"] = index
        return self.curl.getjson("stream/create", params)

    def stream_insert(self, path):
        params = { "path": path }
        return self.curl.getjson("stream/insert", params)


    #     streams = getjson("/stream/list")

    #     eq_(streams, [
    #         ['/newton/prep', 'PrepData'],
    #         ['/newton/raw', 'RawData'],
    #         ['/newton/zzz/rawnotch', 'RawNotchedData'],
    #         ])

    #     streams = getjson("/stream/list?layout=RawData")
    #     eq_(streams, [['/newton/raw', 'RawData']])

    #     streams = getjson("/stream/list?layout=NoSuchLayout")
    #     eq_(streams, [])

    #     with assert_raises(HTTPError) as e:
    #         getjson("/stream/get_metadata?path=foo")
    #     eq_(e.exception.code, 404)

    #     data = getjson("/stream/get_metadata?path=/newton/prep")
    #     eq_(data, {'description': 'The Data', 'v_scale': '1.234'})

    #     data = getjson("/stream/get_metadata?path=/newton/prep"
    #                    "&key=v_scale")
    #     eq_(data, {'v_scale': '1.234'})

    #     data = getjson("/stream/get_metadata?path=/newton/prep"
    #                    "&key=v_scale&key=description")
    #     eq_(data, {'description': 'The Data', 'v_scale': '1.234'})

    #     data = getjson("/stream/get_metadata?path=/newton/prep"
    #                    "&key=v_scale&key=foo")
    #     eq_(data, {'foo': None, 'v_scale': '1.234'})

    #     data = getjson("/stream/get_metadata?path=/newton/prep"
    #                    "&key=foo")
    #     eq_(data, {'foo': None})


    # def test_insert(self):
    #     # invalid path first
    #     with assert_raises(HTTPError) as e:
    #         getjson("/stream/insert?path=/newton/")
    #     eq_(e.exception.code, 404)

    #     # XXX TODO
    #     with assert_raises(HTTPError) as e:
    #         getjson("/stream/insert?path=/newton/prep")
    #     eq_(e.exception.code, 501)



